/*-------------------------------------------------------------------*/
/* CBERROR.C      Juha Kivijrvi                                     */
/*                Timo Kaukoranta                                    */
/*                                                                   */
/* - Calculates average error between codebook and training set.     */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName       "CBERROR"
#define VersionNumber  "Version 0.28"
#define LastUpdated    "7.7.16" /* PF */

/* ----------------------------------------------------------------- */
/*      Changelog:                                                   */
/*                                                                   */
/* PF     0.28  Fixed CI to Two-Ways + use PrintError function       */
/* PF     0.27  Minor quick-n-dirty tuning of CI-index printing      */
/* PF     0.26  Error messages added. QuietLevel=1 option below.     */
/* PF     0.25  Added -g option: global allocation (CI-index)        */
/* PF     0.24  Minor changes in style and QuietLevel                */
/* JK/TK  0.22  Original version from 8.1.97                         */
/*                                                                   */
/* ----------------------------------------------------------------- */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <values.h>

#define  FACTFILE  "cberror.fac"

#include "parametr.c"
#include "cb.h"
#include "file.h"
#include "memctrl.h"
#include "owntypes.h"

#define width(a)   ( (a) > 0  ? (int) log10(a) + 1 : \
                     (a) == 0 ? 1 : (int) log10(-a) +2 )

/*============================  P R I N T  ===============================*/


static void PrintOperatingInfo(char  TSetName[],
                               char  CBName[])
{
  if( Value(QuietLevel) >= 2 )
    {
    printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
    printf("Training Set: %s\n", TSetName);
    printf("Codebook:     %s\n", CBName);
    PrintSelectedOptions();
    printf("\n");
    }
  else if( Value(QuietLevel) >= 1 )
    {
    printf("%-12s ", TSetName);
    printf("%-12s ", CBName);
    }
}


/*-------------------------------------------------------------------*/


void PrintInfo(void)
{
  printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
  printf("This program calculates statistics of a training set and a codebook.\n");
  printf("Usage: %s [%coption] <training-set> <codebook>\n", ProgName, OPTION_SYMBOL);
  printf("For example: %s bridge bridge-km\n", ProgName);
  printf("\n  Options:\n");
  PrintOptions();
  printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
}


/*=========================  M A T H  ======================================*/


static double Average(double* values, int* freqs, int count)
{
  int    i;
  double sum     = 0.0;
  int    freqsum = 0;

  if (count == 0) return 0.0;
  
  for (i=0; i<count; i++)
    {
    sum     += values[i] * freqs[i];
    freqsum += freqs[i];
    }
  
  return (sum / freqsum);
}
  

/*-------------------------------------------------------------------*/


static double StandardDeviation(double* values, 
                                int*    freqs,
                                int     count, 
                                double  average)
{
  int    i;
  double diffsum = 0.0;
  int    freqsum = 0;
  
  if (count == 0) return 0.0;

  for (i=0; i<count; i++)
    {
    diffsum += (values[i] - average) * (values[i] - average) * freqs[i];
    freqsum += freqs[i];
    }
    
  return sqrt(diffsum / freqsum);
}    


/*-------------------------------------------------------------------*/


static double Dissimilarity(TRAININGSET* TS, PARTITIONING* P, int index)
{
  int     i,j;
  double* errors;
  int*    freqs;
  int     count;
  int     ind = 0;
  double  aver;
  double  stdev;

  if (UniqueVectors(P,index) < 2) return 0.0;

  count  = (UniqueVectors(P,index) * (UniqueVectors(P,index) + 1)) / 2;

  errors = allocate( count * sizeof(double) );
  freqs  = allocate( count * sizeof(int) );

  for (i=FirstVector(P,index); !(EndOfPartition(i)); i=NextVector(P,i))
    {
    errors[ind] = 0;
    freqs[ind]  = VectorFreq(TS,i) * (VectorFreq(TS,i)-1) / 2;
    ind++;
    for (j=FirstVector(P,index); i!=j; j=NextVector(P,j))
      {
      errors[ind] = sqrt( (double) VectorDist(Vector(TS,i),
                                              Vector(TS,j),
                                              VectorSize(TS)));
      freqs[ind]  = VectorFreq(TS,i) * VectorFreq(TS,j);
      ind++;
      }
    }

  aver  = Average(errors, freqs, count);
  stdev = StandardDeviation(errors, freqs, count, aver);

  deallocate(errors);
  deallocate(freqs);
  return ( aver < 0.000000001 ? 0.0 : stdev / aver );  
}      


/*-------------------------------------------------------------------*/


static void CalculateMAEandNMatches(TRAININGSET* TS, 
                                    CODEBOOK* CB,
                                    PARTITIONING* P,
                                    double* MAE,
                                    int* nMatches)
{
  int    i,j;
  int    MAEdiff;
  llong  MAEsum    = 0LL;

  *nMatches = 0;
  for (i=0; i<BookSize(TS); i++)
    {
    MAEdiff = 0;
    for (j=0; j<VectorSize(TS); j++)
      {
      MAEdiff += VectorFreq(TS,i) * abs( VectorScalar(TS,i,j) - 
                                         VectorScalar(CB,Map(P,i),j) );
      }
    if (MAEdiff == 0) (*nMatches)++;
    MAEsum += MAEdiff;
    }
  *MAE = (double) MAEsum / ((double)TotalFreq(TS) * (double)VectorSize(TS));
}


/*-------------------------------------------------------------------*/


static void PartitionErrors(TRAININGSET* TS, 
                            CODEBOOK* CB,
                            PARTITIONING* P)
{
  int    i, j;
  llong  PErrorSum = 0LL;
  int    vectorlength;
  double dissim;
  double dissimsum = 0.0;

  printf("\nind   ");
  vectorlength = VectorSize(CB) * (width(CB->MaxValue) + 
                 (Value(NoBlanks) ? 0 : 1)) + 1;

  /* Print heading for the partition data */
  if (vectorlength < 11)
    for (i=0; i<vectorlength; i++) printf(" ");
  else
    {
    printf("codevector");
    for (i=10; i<vectorlength; i++) printf(" ");
    }
  printf(" %5s %5s %8s %9s %9s\n", 
         "freq", "uniq", "sqerror", "MSE", "dissim");
 
  /* Process the clusters one by one */
  for (i=0; i<BookSize(CB); i++)
    {
    /* Go through the vectors in cluster i */
    PErrorSum = 0LL;
    for (j=FirstVector(P,i); !(EndOfPartition(j)); j=NextVector(P,j))
      PErrorSum += VectorDist( Vector(TS,j), Vector(CB,i), VectorSize(TS) ) *
                   VectorFreq(TS,j);
        
    /* Print partition index */
    printf("%3d | ", i+1);

    /* Print the centroid of cluster i */
    for (j=0; j<VectorSize(CB); j++)
      {
      printf("%*d", width(CB->MaxValue), VectorScalar(CB,i,j));
      if (!(Value(NoBlanks))) printf(" ");
      }                         
    if (Value(NoBlanks)) printf(" ");

    /* Print the calculated statistics for cluster i */
    printf("| %5d %5d %8d ",CCFreq(P,i), UniqueVectors(P,i), (int)PErrorSum);
    printf("%9.4f ", ( CCFreq(P,i) == 0 ? 0.0 : (double)PErrorSum /
                        ((double)CCFreq(P,i) * VectorSize(CB)) ) ); 
    dissim = Dissimilarity(TS,P,i);
    dissimsum += dissim * CCFreq(P, i);;
    printf("%9.4f\n", dissim);
    } 

  printf("Average dissimilarity = %f\n", dissimsum / TotalFreq(TS));
}


/*-------------------------------------------------------------------*/
/* CI1 index for global allocation when comparing two CBs            */
/* Not used anymore                                                  */
/*-------------------------------------------------------------------*/


static void GlobalAllocations(TRAININGSET* TS, 
                              CODEBOOK* CB,
                              PARTITIONING* P)
{
  int  i, count;
  int  total=0;
  int  zeros=0, ones=0, plenty=0;

  if( Value(GlobalAllocation)==NO ) return;  /* Quick-and-dirty */

  /* First sketch to see if the partition counts give the answer directly. */
  for (i=0; i<BookSize(CB); i++)
    {
    count = UniqueVectors(P,i);
    if(count==0) zeros++; 
    if(count==1) ones++;
    if(count>=2) plenty++;
    if(Value(QuietLevel)>=3) printf("%i ", count);
    total += count;
    }

  if(Value(QuietLevel)>=3) printf("\n");
  if( Value(QuietLevel) >= 2 )
    {
    printf("\nGlobal matching result:\n");
    printf("  No matches:   %i \n", zeros);
    printf("  One match:    %i \n", ones);
    printf("  Two or more:  %i \n", plenty);
    if(ones==BookSize(CB)) 
       printf("CORRECT MATCH. \n");
    else
       printf("The first file needs %i swaps to match the second one.\n", zeros);
    } 
  else if( Value(QuietLevel) >= 1 )
    {
    printf("  CI=%i  ", zeros);
    }

  /* Validity checkings. Note: Frequencies are ignored!!! */
  if((zeros+ones+plenty)!=BookSize(CB)) printf("Warning: incorrect counts in Global Allocation!\n");
  if( total != BookSize(CB))
    {
    printf("\nNOTE!!! Cluster count %i dosn't match ", total);
    printf("the size of the second file %i. ", BookSize(CB));
    printf("\nThe analysis might therefore be meaningless.\n");
    }
}


/*=========================  CODEBOOK HANDLING  ============================*/


static void CodeTSWithCB(TRAININGSET*  TS, 
                         CODEBOOK*     CB, 
                         PARTITIONING* P,
                         double        MSE)
{
  double MAE = 0.0;
  int    ci=0, nMatches=0;

  CalculateMAEandNMatches(TS, CB, P, &MAE, &nMatches);
  if( Value(GlobalAllocation) ) 
    {
    ci = CentroidIndex(TS, CB);
    }

  if( Value(QuietLevel) >= 2 )
    {
    printf("MSE Distortion = %s\n", FormattedValue(PrintableError(MSE,CB)));
    printf("MAE Distortion = %s\n", FormattedValue(PrintableError(MAE,CB)));
    printf("Equal vectors  = %d\n",   nMatches);
    if( Value(GlobalAllocation) )  printf("Centroid index = %i\n", ci);
    printf("\n");
    }
  else if( Value(QuietLevel) >= 1 )
    {
    printf("%s   ", FormattedValue(PrintableError(MSE,CB)));
    printf("%s   ", FormattedValue(PrintableError(MAE,CB)));
    printf("%d   ",  nMatches);
    if( Value(GlobalAllocation) )  printf("%i   ", ci);
    printf("\n");
    }
  else
    {
    printf("%9.4f  ", PrintableError(MSE,CB));
    }

  /* Print errors for each partition separately */
  if( Value(PartitionErrors) )
    {
    PartitionErrors(TS, CB, P);
    }

}


/* ----------------------------------------------------------------- */


static void CodeBothWays(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* P)
{
  double MSE1 = 0.0, MSE2 = 0.0;
  double MAE1 = 0.0, MAE2 = 0.0;
  int    nMatches1 = 0, nMatches2 = 0;

  MSE1 = GenerateOptimalPartitioning(TS, CB, P);

  CalculateMAEandNMatches(TS, CB, P, &MAE1, &nMatches1);

  MSE2 = GenerateOptimalPartitioning(CB, TS, P);

  CalculateMAEandNMatches(CB, TS, P, &MAE2, &nMatches2);

  if( Value(QuietLevel) >= 2 )
    {
    printf("                  TS->CB    CB->TS    Average\n");
    printf("MSE Distortion = %9.4f %9.4f %9.4f\n", 
            PrintableError(MSE1,CB), 
            PrintableError(MSE2,CB),
            PrintableError((MSE1+MSE2)/2.0,CB));
    printf("MAE Distortion = %9.4f %9.4f %9.4f\n", 
            PrintableError(MAE1,CB), 
            PrintableError(MAE2,CB),
            PrintableError((MAE1+MAE2)/2.0,CB));
    printf("Equal vectors  = %9d %9d\n", nMatches1, nMatches2);
    }
  else if( Value(QuietLevel) >= 1 )
    {
    printf("%9.4f %9.4f %9.4f\n", PrintableError(MSE1,CB),             
                                  PrintableError(MSE2,CB),             
                                  PrintableError((MSE1+MSE2)/2.0,CB)); 
            
    }
  else
    {
    printf("%9.4f  ", PrintableError((MSE1+MSE2)/2.0,CB));
    }
}


/* ----------------------------------------------------------------- */


static void DoCalculations(TRAININGSET*  TS, 
                           CODEBOOK*     CB, 
                           PARTITIONING* P,
                           CBFILETYPE    filetype)
{
  double     MSEv;
  ERRORFTYPE errorf;

  if (Value(ErrorFunction) == MSE_D)
    {
    if ( filetype == CBFILE || filetype == TSFILE )
      {
      if( Value(DoubleWayError) == NO )
        {
        MSEv = GenerateOptimalPartitioning(TS, CB, P);
        CodeTSWithCB(TS, CB, P, MSEv);
        }
      else
        {
        if ( BookSize(TS) == BookSize(CB) )
          CodeBothWays(TS, CB, P);
        else
          printf("ERROR: TrainingSet and Codebook sizes differ.\n");
        }
      }
    else
      {
      GenerateOptimalCodebook(TS, CB, P);
      MSEv = AverageErrorForSolution(TS, CB, P, MSE);
      CodeTSWithCB(TS, CB, P, MSEv);
      }
    }
  else 
    { 
    switch (Value(ErrorFunction))
        {
        case SC_D:      errorf = SC;          break;
        case JACCARD_D: errorf = JACCARD;     break;
        case DICE_D:    errorf = DICE;        break;
        default:        errorf = MSE;
        }
    if ( filetype == CBFILE || filetype == TSFILE )
      {
      MSEv = AverageErrorCB(TS, CB, errorf);
      }
    else
      {
      MSEv = AverageErrorForPartitioning(TS, P, errorf);
      }
    printf("Distortion = %9.4f\n", MSEv);
    }

}


/* ================================================================= */


static void Initialize(TRAININGSET*  TS, 
                       CODEBOOK*     CB, 
                       PARTITIONING* P,
                       CBFILETYPE    filetype)
{
  int ndups;

  if( (ndups = DuplicatesInCodebook(TS)) > 0 )
    {
    printf("\nWARNING: TrainingSet contains %i duplicates.\n\n",
           ndups);
    }

  if (filetype == CBFILE || filetype == TSFILE)
    {
    CreateNewPartitioning(P, TS, BookSize(CB));
    if( (ndups = DuplicatesInCodebook(CB)) > 0 )
      {
      printf("\nWARNING: Codebook contains %i duplicates.\n\n",ndups);
      }
    }
  else
    {
    CreateNewCodebook(CB, P->PartitionCount, TS);
    }

}


/* ================================================================= */


static void ShutDown(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* P)
{
  FreeCodebook(TS);
  FreeCodebook(CB);
  FreePartitioning(P);
}


/*===========================  M A I N  ==============================*/


int main(int argc, char* argv[])
{
  CODEBOOK      CB;
  TRAININGSET   TS;
  PARTITIONING  P;
  char          TSName[80];
  char          CBName[80];
  CBFILETYPE    ft1, ft2;

  ParameterInfo paraminfo[2] =  { { TSName, "", 0, INFILE  }, 
                                  { CBName, "", 0, INFILE  } };
                               
  ParseParameters( argc, argv, 2, paraminfo);
  ft1 = DetermineCBFileTypeConsideringOrder(TSName, 12);
  ft2 = DetermineCBFileTypeConsideringOrder(CBName, 231);
  if ( (ft1 == CBFILE) && (ft2 == PAFILE))
     {
     printf("\nERROR: 1st file should be TS since 2nd is PA: s%s, %s \n", TSName, CBName);
     exit(-1);
     }
  if ( ft1 == NOTFOUND )
     {
     printf("\nERROR: Training set file not found: %s \n", TSName);
     exit(-1);
     }
  if ( ft2 == NOTFOUND )
     {
     printf("\nERROR: Second file name not found: %s \n", CBName);
     exit(-1);
     }

  PrintOperatingInfo(TSName, CBName);
  ReadTrainingSet(TSName, &TS);
  (ft2 == PAFILE) ? ReadPartitioning(CBName, &P, &TS) :
                    ReadCodebook(CBName, &CB);
  Initialize(&TS, &CB, &P, ft2);

  DoCalculations(&TS, &CB, &P, ft2);

  ShutDown(&TS, &CB, &P);
  checkmemory();
  return( 0 );
}

